/*~~~~~~~~~~~~~~ mwplight - The MegaWave2 Light Preprocessor ~~~~~~~~~~~~~~~~~~~

 Generate the M-file (modfile : source of modified module)

 Author : Jacques Froment
 Date : 2007
 Version : 1.0
 Versions history :
   0.1 (August 2005, JF) initial internal release
   1.0 (April 2007, JF) final revision for external release
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~  This file is part of the MegaWave2 light preprocessor ~~~~~~~~~~~~
MegaWave2 is a "soft-publication" for the scientific community. It has
been developed for research purposes and it comes without any warranty.
The last version is available at http://www.cmla.ens-cachan.fr/Cmla/Megawave
CMLA, Ecole Normale Superieure de Cachan, 61 av. du President Wilson,
      94235 Cachan cedex, France. Email: megawave@cmla.ens-cachan.fr 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/* TO DO :
   add the ANSI declaration (#ifdef __STDC__ ...) for the main function, so that
   we can prototype it using the ANSI syntax (change call to WriteFuncPrototype()
   in mwpl_afile.c and see note in WriteFuncPrototype()
*/

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>

#include "mwpl_main.h"

/*~~~ write the header ~~~
*/

#ifdef __STDC__
void writeMheader(void)
#else
void writeMheader()
#endif
{
  fprintf(fm,"/* This M-file has been generated by the MegaWave2 Light Preprocessor.\n");
  fprintf(fm,"   It contains the source of the modified module '%s' (to be archived in library).\n",module_name);
  fprintf(fm,"*/\n\n");
  
  /* Write linemarker preprocessing directive so that, in case of error during the compilation,
     the C compiler will write the name of the module file and the line number where the error
     is, instead of the name and line number of the M-file.
     Beware :
     1/ this may work for gcc compilers only (not checked for others) ;
     2/ in order to get the right line number, do not change the source code after the linemarker.
        So, do not put comments.
	Also, when ANSI declaration will be added, this linemarker will have to be adapted (by e.g.
	adding a new one after the declaration to re-synchronize line numbers following the
	example of printfslocation()).
  */
  fprintf(fm,"# 1 \"%s\"\n",module_file);
}  

/*~~~ write additional text for VarFunc <f> ~~~
*/

#ifdef __STDC__
void writeMVarFunc(VarFunc *f)
#else
void writeMVarFunc(f)
VarFunc *f;
#endif
{
  /* Function declarations (but the main one) are local to the module file 
     (not to be archived in library); therefore the C storage keyword "static"
     has to be added in front of function declarations.
  */

  if (ISCI_FUNCDECL(f)&&(C->mfunc!=f)&&(f->v->Cstorage[0]=='\0'))
      fprintf(fm,"static ");

  /* TO DO : in case of the main function declaration, replace it
             by the content of <protobuf> (K&R and ANSI declarations).
	     See setprotobuf() for more comments.
  */
}


/*~~~ write the body ~~~
*/

#ifdef __STDC__
void writeMbody(void)
#else
void writeMbody()
#endif
{
  int l;
  long n;
  VarFunc *f;
  
  f=C->varfunc;
  if (!f) Error("[writeMbody] C body not parsed !");

  fs = fopen(module_file,"r");
  if (!fs) Error("Cannot reopen file \"%s\" for reading",module_file);
  n=0;

  while ((l=getc(fs))!=EOF) 
    {
      putc(l,fm);
      n++;
      if (f && (n==((f->l0)-1))) 
	{
	  writeMVarFunc(f);
	  f=f->next;
	}
    }

  fclose(fs); fs=NULL;
} 


/*~~~ main entry : generate M-file ~~~
*/

#ifdef __STDC__
void genMfile(void)
#else
void genMfile()
#endif
{
  fm = fopen(Mfile,"w");
  if (!fm) Error("Cannot open file \"%s\" for writing",Mfile);
  
  writeMheader();
  writeMbody();

  fclose(fm);
}
