/*~~~~~~~~~~~~~~ mwplight - The MegaWave2 Light Preprocessor ~~~~~~~~~~~~~~~~~~~

 Generate the T-file (texfile: module's documentation in LaTeX)

 Author : Jacques Froment
 Date : 2007
 Version : 1.0
 Versions history :
   0.1 (August 2005, JF) initial internal release
   1.0 (April 2007, JF) final revision, ready for external release
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~  This file is part of the MegaWave2 light preprocessor ~~~~~~~~~~~~
MegaWave2 is a "soft-publication" for the scientific community. It has
been developed for research purposes and it comes without any warranty.
The last version is available at http://www.cmla.ens-cachan.fr/Cmla/Megawave
CMLA, Ecole Normale Superieure de Cachan, 61 av. du President Wilson,
      94235 Cachan cedex, France. Email: megawave@cmla.ens-cachan.fr 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include "mwpl_main.h"

/*~~~ write the header and first LaTeX macros ~~~
*/

#ifdef __STDC__
void writeTheader(void)
#else
void writeTheader()
#endif
{
  /* Comments which may be used by a shell to make the whole doc */
  fprintf(ft, "%%*** Group=%s\n",group_name);
  fprintf(ft, "%%*** Name=%s\n\n",module_name);

  fprintf(ft,"%% This T-file has been generated by the MegaWave2 Light Preprocessor.\n");
  fprintf(ft,"%% It contains the LaTeX document skeleton (doc) of the module '%s'.\n\n",module_name);

  /* Header for the page */
  fprintf(ft,"\\markboth");
  fprinttex(ft,"{{\\em %T} \\hfill MegaWave2 User's Modules Library \\hfill {\\bf %T} \\hspace{1cm}}",
	    group_name,module_name);
  fprinttex(ft,"{{\\em %T} \\hfill MegaWave2 User's Modules Library \\hfill {\\bf %T} \\hspace{1cm}}\n\n",
	    group_name,module_name);
  
  /* Label for the index */
  fprintf(ft,"\\label{%s}\n\n",module_name);

  /* Index entry  */
  fprinttex(ft, "\\index{%T@{\\tt %T}}\n\n",module_name,module_name);  

  /* Name */
  fprinttex(ft, "\\Name{%T}{%T}\n\n",module_name,H->Function);
}  

/*~~~ write the Synopsis LaTeX macro.
      Code similar to the usage one in the A-file (see writeusage() in mwpl_afile)
  ~~~
*/

#ifdef __STDC__
void writeSynopsis(void)
#else
void writeSynopsis()
#endif
{ 
  Arg *a;

  fprinttex(ft, "\\Synopsis{%T}{",module_name);
   /* Options */
  if (H->NbOption>0)
    for (a=H->usage;a;a=a->next)
      if (ISARG_OPTION(a))
	{
	  fprintf(ft," [-%c", a->Flag);
	  if (a->H_id[0]!='\0') 
	    fprinttex(ft," {\\em %T}] ", a->H_id);
	  else 
	    fprinttex(ft,"] ");
        }
  /* Needed arg */
  if (H->NbNeededArg>0)
    for (a=H->usage;a;a=a->next)
      if (ISARG_NEEDED(a))
	{
	  if (!(ISARG_SCALAR(a) && ISARG_OUTPUT(a)))
	    fprinttex(ft,"{\\em %T} ", a->H_id);
	  else
	    /* In case of a scalar output needed argument, something (as .) has to be
	       given in the command line to avoid argument shifting.
	    */
	    fprinttex(ft,"{\\em .} ");
	}
  /* Optional argument */
  if (H->NbOptionArg>0)
    {
     /* BEWARE : see if it would be useful to check if the following test
	 is never verified for all optional arg, in order not to write []. 
      */
      fprinttex(ft, "[");
      for (a=H->usage;a;a=a->next)
	if (ISARG_OPTARG(a))
	  {
	    if (!(ISARG_SCALAR(a) && ISARG_OUTPUT(a)))
	      fprinttex(ft,"{\\em %T} ", a->H_id);
	  }
      fprinttex(ft,"] ");
    }

  /* Variable arguments */
  if (H->NbVarArg>0)
    for (a=H->usage;a;a=a->next)
      if (ISARG_VARIABLE(a))
	{
	  if (!(ISARG_SCALAR(a) && ISARG_OUTPUT(a)))
	    fprinttex(ft,"{\\em ...}");
	}
  
  fprinttex(ft, "}\n\n");
}


/*~~~  
  Print in the T-file the interval and the default
  value of the argument's value, if applicable.
  Code similar to print_interval_and_defval() (see mwpl_afile)
  ~~~
*/

#ifdef __STDC__
void  printtex_interval_and_defval(Arg *a)
#else
void  printtex_interval_and_defval(a)
Arg *a;
#endif

{
  char t_m[3], t_M[3], fc[256];
  char A,B;

  if (!(ISARG_INTERVAL(a) || ISARG_DEFAULT(a))) return;

  fprintf(ft, " (");
  if (ISARG_INTERVAL(a))
    {
      switch (a->ICtype) 
	{
	case CLOSED :
	  A='['; B=']';
	  break;
	case MAX_EXCLUDED :
	  A='['; B='[';
	  break;
	case MIN_EXCLUDED :
	  A=']'; B=']';
	  break;
	case OPEN :
	  A=']'; B='[';
	  break;
	default :
	  Error("[printtex_interval_and_defval] Unexpected error for C_id=\"%s\" : illegal ICtype=%d",a->C_id,a->ICtype);
	  break;
	}
      fprinttex(ft, "in %c%T,%T%c",A,a->Min,a->Max,B);
      if (ISARG_DEFAULT(a)) fprintf(ft, ", ");
    }
  if (ISARG_DEFAULT(a)) fprintf(ft, "default %s",a->Val);
  fprintf(ft, ")");
}

/*~~~ 
  write Arguments.
  Beware, with LaTeX 2e do not use macro \Argument but plain text in order to allow \verb.
  Code similar to the usage one in the A-file (see writeusage() in mwpl_afile)
  ~~~
*/

#ifdef __STDC__
void writeArguments(void)
#else
void writeArguments()
#endif
{ 
  Arg *a;

  fprinttex(ft, "\n% --- ARGUMENTS : BEGIN --- \n");
  fprinttex(ft, "\\nopagebreak\n");
  fprinttex(ft, "\\vspace{-5mm}\n");  
  fprinttex(ft, "\\nopagebreak\n");
  fprinttex(ft, "\\begin{quotation}\n");

  /* Options */
  if (H->NbOption>0)
    for (a=H->usage;a;a=a->next)
      if (ISARG_OPTION(a))
	{
	  fprintf(ft,"-%c ", a->Flag);
	  if (a->H_id[0]!='\0') fprinttex(ft, "{\\em %T}",a->H_id);
	  printtex_interval_and_defval(a);
	  if (ISARG_SCALAR(a) && ISARG_OUTPUT(a))
	    fprinttex(ft," screen output");
          fprinttex(ft," : %T\n\n", getprintfstring(a->Cmt));
	}

  /* Needed arg */
  if (H->NbNeededArg>0)
    for (a=H->usage;a;a=a->next)
      if (ISARG_NEEDED(a))
	{
	  if (!(ISARG_SCALAR(a) && ISARG_OUTPUT(a)))
	    {
	      fprinttex(ft, "{\\em %T}",a->H_id);
	      printtex_interval_and_defval(a);
	      fprinttex(ft," : %T\n\n", getprintfstring(a->Cmt));
	    }
	  else
	    fprinttex(ft,". (screen output) : %T\n\n", getprintfstring(a->Cmt));
	}  

 /* Optional arguments */
  if (H->NbOptionArg>0)
    for (a=H->usage;a;a=a->next)
      if (ISARG_OPTARG(a))
	{
	  if (!(ISARG_SCALAR(a) && ISARG_OUTPUT(a)))
	    {
	      fprinttex(ft, "{\\em %T}",a->H_id);
	      printtex_interval_and_defval(a);
	      fprinttex(ft," : %T\n\n", getprintfstring(a->Cmt));
	    }
	  else
	    fprinttex(ft,"screen output : %T\n\n", getprintfstring(a->Cmt));
	}  
  
  /* Variable arg */
  if (H->NbVarArg>0)
    for (a=H->usage;a;a=a->next)
      if (ISARG_VARIABLE(a))
	{
	  if (!(ISARG_SCALAR(a) && ISARG_OUTPUT(a)))
	    fprinttex(ft,"... : %T\n\n",getprintfstring(a->Cmt));
	  else
	    fprinttex(ft,"screen output... : %T\n\n",getprintfstring(a->Cmt));
	}
  
  fprinttex(ft, "\\end{quotation}\n");
  fprinttex(ft, "\\Next\n");
  fprinttex(ft, "% --- ARGUMENTS : END --- \n\n");

}


/*~~~ 
  write Summary.
  Beware, with LaTeX 2e do not use macro \Argument but plain text in order to allow \verb.
  Code similar to the one in setprotobuf() (See mwpl_ifile.c).
  As with setprotobuf, both K&R and ANSI declaration should be implemented.
  ~~~
*/

#ifdef __STDC__
void writeSummary(void)
#else
void writeSummary()
#endif
{ 
  VarFunc *f;
  Variable *p;
  
  fprinttex(ft, "\n% --- SUMMARY : BEGIN --- \n");
  fprinttex(ft, "\\Mark{\\Large\\bf  Function Summary} \\nopagebreak\\bigskip\n");
  fprinttex(ft, "\\nopagebreak\n\n");

   /* Function name and arguments list */
  f=C->mfunc;
  fprinttex(ft,"%T %T ( ",f->v->Ftype,f->v->Name);
  for (p=f->param; p; p=p->next)
    {
      if (p!=f->param) fprintf(ft," , ");
      fprinttex(ft,"%T",p->Name);
    } 
  fprintf(ft," )\n\n");
  
  /* Type of arguments */
  for (p=f->param; p; p=p->next)
    fprinttex(ft,"%T %T ;\n\n",p->Ftype,p->Name);

  fprinttex(ft, "\\Next\n");
  fprinttex(ft, "% --- SUMMARY : END --- \n\n");
}


/*~~~ 
  write Description.
  Beware, with LaTeX 2e do not use macro \Argument but plain text in order to allow \verb.
  ~~~
*/

#ifdef __STDC__
void writeDescription(void)
#else
void writeDescription()
#endif

{ 
  fprinttex(ft, "\n% --- DESCRIPTION : BEGIN --- \n");
  fprinttex(ft, "\\Mark{\\Large\\bf  Description} \\nopagebreak\\bigskip\n");
  fprinttex(ft, "\\nopagebreak\n\n");
  fprinttex(ft, "\\nopagebreak\n");
  fprintf(ft,"\\input{src/%s.tex}\n\n",module_name);  
  fprinttex(ft, "\\Next\n");
  fprinttex(ft, "% --- DESCRIPTION : END --- \n\n");
}

/*~~~ 
  write the foot (last LaTeX macros).
  ~~~
*/

#ifdef __STDC__
void writeTfoot(void)
#else
void writeTfoot()
#endif

{ 
  char User_DepFile[BUFSIZ];
  char *lastmodifdate;
  char Auth[TREESTRSIZE], Lab[TREESTRSIZE];
  char An0[5],An1[5];

  /* See also */
  sprintf(User_DepFile,"%s.dep",module_name);
  fprintf(ft,"\\input{obj/DEPENDENCIES/%s}\n\n",User_DepFile);

  /* Get date of last modification of the module file */
  lastmodifdate=ctime(&(module_fstat.st_mtime));

  /* Version */
  fprinttex(ft, "\\Version{%T}{%s}\n",H->Version,lastmodifdate);

  /* Author */
  strcpy(An0,"1993");
  strncpy(An1,&lastmodifdate[strlen(lastmodifdate)-5],4);
  An1[4]='\0';
  
  if (H->Labo[0]!='\0') strcpy(Lab,getprintfstring(H->Labo));
  else strcpy(Lab,"CMLA, ENS Cachan, 94235 Cachan cedex, France");
  
  strcpy(Auth,getprintfstring(H->Author));
  if (strcmp(An0,An1) == 0)
    fprinttex(ft, "\\Author{%s}{%s}{%T}\n",Auth,An0,Lab);
  else
    fprinttex(ft, "\\Author{%s}{%s-%s}{%T}\n",Auth,An0,An1,Lab);

}


/*~~~ main entry : generate T-file ~~~
*/

#ifdef __STDC__
void genTfile(void)
#else
void genTfile()
#endif
{ 

  if (Tfile[0]=='\0') sprintf(Tfile,"%s.tex",H->Name);
  
  if ((ft = fopen(Tfile, "w")) == NULL) 
    Error("Cannot open T-file '%s' for writing",Tfile);    
  
  writeTheader();
  writeSynopsis();
  writeArguments();
  writeSummary();
  writeDescription();
  writeTfoot();

  fclose(ft);
}
